
#ifndef CPP_START_ASPECT_H
#define CPP_START_ASPECT_H

#define HAS_MEMBER(member)\
template<typename T, typename... Args>\
class HasMember_##member\
{\
private:\
    template<typename U>\
    static auto check(int) -> decltype(std::declval<U>().member(std::declval<Args>()...), std::true_type());\
\
    template<typename U>\
    static std::false_type check(...);\
\
public:\
    enum { value = std::is_same<decltype(check<T>(0)), std::true_type>::value };\
};\


HAS_MEMBER(before)

HAS_MEMBER(after)

#include "../base/non_copy_able.h"

template<typename Func, typename... Args>
class Aspect : NonCopyAble {
private:
    Func mFunc;

public:
    explicit Aspect(Func &&f) : mFunc(std::forward<Func>(f)) {
    }

    template<typename T>
    typename std::enable_if<HasMember_before<T, Args...>::value && HasMember_after<T, Args...>::value>::type
    invoke(Args &&... args, T &&aspect) {
        aspect.before(std::forward<Args>(args)...);
        // TODO 推断返回类型，若有返回值则将返回值作为后置处理的参数
        mFunc(std::forward<Args>(args)...);
        aspect.after(std::forward<Args>(args)...);
    }

    template<typename T>
    typename std::enable_if<HasMember_before<T, Args...>::value && !HasMember_after<T, Args...>::value>::type
    invoke(Args &&... args, T &&aspect) {
        aspect.before(std::forward<Args>(args)...);
        mFunc(std::forward<Args>(args)...);
    }

    template<typename T>
    typename std::enable_if<!HasMember_before<T, Args...>::value && HasMember_after<T, Args...>::value
    >::type invoke(Args &&... args, T &&aspect) {
        mFunc(std::forward<Args>(args)...);
        aspect.after(std::forward<Args>(args)...);
    }

    template<typename Head, typename... Tail>
    void invoke(Args... args, Head &&headAspect, Tail &&... tailAspect) {
        headAspect.before(std::forward<Args>(args)...);
        invoke(std::forward<Args>(args)..., std::forward<Tail>(tailAspect)...);
        headAspect.after(std::forward<Args>(args)...);
    }
};

/**
 * 辅助函数，简化调用
 * @tparam AOP 切点
 * @tparam Func 切点方法类型
 * @tparam Args 方法参数类型
 * @param func 切点方法
 * @param args 方法参数
 */
template<typename... AOP, typename Func, typename... Args>
void aop(Func &&func, Args &&... args) {
    Aspect<Func, Args...> asp(std::forward<Func>(func));
    asp.invoke(std::forward<Args>(args)..., AOP()...);
}

#endif //CPP_START_ASPECT_H
