
#ifndef CPP_START_MESSAGE_BUS_H
#define CPP_START_MESSAGE_BUS_H

#include <string>
#include <functional>
#include <map>
#include "../base/any.h"
#include "../base/function_traits.h"
#include "../base/non_copy_able.h"

using namespace std;

class MessageBus : public NonCopyAble {
public:
    //注册消息
    template<typename F>
    void Attach(F &&f, const string &strTopic = "") {
        auto func = to_function(std::forward<F>(f));
        Add(strTopic, std::move(func));
    }

    //发送消息
    template<typename R>
    //无参数
    void SendReq(const string &strTopic = "") {
        //std::function  函数对象类，包装了满足该函数模板类型的任意函数对象，这些目标实体包括普通函数、Lambda表达式、函数指针、模函数、类成员函数、类静态函数以及其它的函数对象
        using function_type = std::function<R()>;
        string strMsgType = strTopic + string("##") + typeid(function_type).name();

        std::pair<std::multimap<string, Any>::iterator, std::multimap<string, Any>::iterator> range = m_map.equal_range(
                strMsgType);
        for (Iterator it = range.first; it != range.second; ++it) {
            auto f = it->second.AnyCast<function_type>();
            f();
        }
    }

    template<typename R, typename... Args>
    //有参数
    void SendReq(Args &&... args, const string &strTopic = "") {
        using function_type = std::function<R(Args...)>;
        string strMsgType = strTopic + string("##") + typeid(function_type).name();
        cout << "send Msg: " << strMsgType << endl;
        auto range = m_map.equal_range(strMsgType);
        for (Iterator it = range.first; it != range.second; ++it) {
            auto f = it->second.AnyCast<function_type>();
            f(std::forward<Args>(args)...);
        }
    }

    //移除某个主题，需要主题和消息类型
    template<typename R, typename... Args>
    void Remove(const string &strTopic = "") {
        using function_type = std::function<R(Args...)>;
        string strMsgType = strTopic + string("##") + typeid(function_type).name();
        cout << "remove message: " << strMsgType << endl;
        int count = m_map.count(strMsgType);
        auto range = m_map.equal_range(strMsgType);
        m_map.erase(range.first, range.second);
    }

private:
    template<typename F>
    void Add(const string &strTopic, F &&f) {
        string strMsgType = strTopic + string("##") + typeid(F).name();//主题

        //emplace  插入操作，相比inset，其能就地通过参数构造对象，不需要拷贝或者移动内存，使容器插入元素的性能得到进一步提升。
        m_map.emplace(std::move(strMsgType), std::forward<F>(f));

        cout << "register message: " << strMsgType << endl;
    }

private:

    std::multimap<string, Any> m_map; //存储注册的消息
    typedef std::multimap<string, Any>::iterator Iterator;

};

#endif //CPP_START_MESSAGE_BUS_H
